# -*- coding: utf8 -*-

# Copyright 2024 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .aggregator import KSAggregator
from .eigensolver import EigenSolver
from .fluid import UnmatchedFluid, MatchedIsaFluid
from .pk import Pk
from .nipk import Nipk

def init_pypk(cfg):
    """Initialize PyPK
    """
    # Fluid type
    if cfg['fluid'] == 'unmatched':
        n_speed = len(cfg['u_idx']) # number of airspeed test points
        flu = UnmatchedFluid(cfg['rho_inf'], cfg['u_idx'], cfg['l_ref'], cfg['f_ref'], cfg['mu'])
    elif cfg['fluid'] == 'matched_isa':
        n_speed = len(cfg['alt']) # number of airspeed test points
        flu = MatchedIsaFluid(cfg['alt'])
    else:
        raise RuntimeError('"fluid" entry must be "unmatched" or "matched_isa"!\n')
    # Eigen solver
    eig = EigenSolver(cfg['l_ref'], cfg['n_modes'])
    # Aggregator
    agg = KSAggregator(cfg['rho_ks'])
    # p-k method
    if cfg['method'] == 'pk':
        sol = Pk(flu, eig, agg, cfg['mach'], cfg['k_ref'], cfg['l_ref'], cfg['vrb'], cfg['n_modes'], n_speed, tol=1e-3)
    elif cfg['method'] == 'nipk':
        sol = Nipk(flu, eig, agg, cfg['mach'], cfg['k_ref'], cfg['l_ref'], cfg['vrb'], cfg['n_modes'], n_speed)
    else:
        raise RuntimeError('"method" entry must be "pk" or "nipk"!\n')
    return sol
