# -*- coding: utf8 -*-

# Copyright 2024 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import scipy.linalg as spla

class EigenSolver:
    """Eigen solver
    """
    def __init__(self, l, n):
        self._l = l # reference length
        self._n = n # number of modes

    def compute(self, rho, u, m, k, q):
        """Compute eigenvalues and eigenvectors
        rho : fuid density
        u : freestream velocity
        m : mass matrix
        k : stiffness matrix
        q : aerodynamic matrix
        """
        p, v = spla.eig(k - 0.5 * rho * u**2 * q, -u**2 / self._l**2 * m) # det((p*u/l)^2*M + K - q_inf*Q) = 0
        p = np.sqrt(p)
        # Sort
        p = np.where(np.imag(p) < 0, -p, p) # change sign of p if imag(p) < 0
        srt = np.imag(p).argsort()
        # Normalize (optional since scipy does it)
        for j in range(self._n):
            v[:,j] /= spla.norm(v[:,j])
        return p[srt].T, v[:,srt]

    def compute_gradients(self, rho, u, m, k, q, p, v, dm, dk, dq):
        """Compute the derivative of an eigenvalue problem
        rho : fuid density
        u : freestream velocity
        m : mass matrix
        k : stiffness matrix
        q : aerodynamic matrix
        p : eigenvalue solution
        v : eigenvector solution
        m : derivative of mass matrix
        k : derivative of stiffness matrix
        q : derivative of aerodynamic matrix
        """
        a = np.zeros((self._n + 1, self._n + 1), dtype=complex)
        b = np.zeros((self._n + 1, 1), dtype=complex)
        a[:-1,0] = -2 * p * u**2 / self._l**2 * m @ v
        a[:-1,1:] = p**2 * u**2 / self._l**2 * m + k - 0.5 * rho * u**2 * q
        a[-1,0] = 0
        a[-1,1:] = v.T
        b[:-1,0] = (p**2 * u**2 / self._l**2 * dm + dk - 0.5 * rho * u**2 * dq) @ v
        b[-1,0] = 0
        x = spla.lu_solve((spla.lu_factor(a)), b)
        return x[0,0]
