# -*- coding: utf8 -*-

# Copyright 2024 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

class Solution:
    """Base class for solving the aeroelastic equation
    """
    def __init__(self, flu, eig, agg, mach, k_ref, l_ref, vrb, n_modes, n_speed):
        self.name = None # solver name
        self._flu = flu # fluid type
        self._eig = eig # eigen solver
        self._agg = agg # aggregator
        self._mach = mach # Mach number
        self._kref = k_ref # reference reduced frequencies
        self._lref = l_ref # reference length
        self._v = vrb # verbosity level
        # Sizes
        self.n_modes = n_modes # number of modes
        self.n_freqs = len(k_ref) # number of reference reduced frequencies
        self.n_speed = n_speed # number of airspeed test points
        # Outputs
        self.rho = np.zeros(self.n_speed) # density
        self.speed = np.zeros(self.n_speed) # velocity
        self.freq = np.zeros((self.n_speed, self.n_modes)) # frequency
        self.damp = np.zeros((self.n_speed, self.n_modes)) # damping
        self.damp_agg = 0 # aggregated damping
        # Derivatives
        self.damp_m = np.zeros(self.n_modes * self.n_modes) # derivative of aggregated damping wrt mass matrix
        self.damp_k = np.zeros(self.n_modes * self.n_modes) # derivative of aggregated damping wrt stiffness matrix
        self.damp_qre = np.zeros(self.n_freqs * self.n_modes * self.n_modes) # derivative of aggregated damping wrt real aero forces
        self.damp_qim = np.zeros(self.n_freqs * self.n_modes * self.n_modes) # derivative of aggregated damping wrt imag aero forces

    def set_matrices(self, m, k, q):
        """Set structural mass and stiffness, and aerodynamic forces matrices
        """
        self._m = m # mass matrix
        self._k = k # stiffness matrix
        self._q = q # GAF matrices at reference reduced frequencies

    def compute(self):
        raise NotImplementedError()

    def compute_gradients(self):
        raise NotImplementedError()

    def find_flutter(self):
        """Find the flutter speed and display it
        """
        found = False
        self.f_speed = np.inf
        self.f_mode = np.inf
        for ivel in range(len(self.speed) - 1):
            if (found):
                break
            for imod in range(self.n_modes):
                if (self.damp[ivel, imod] > 0 or self.damp[ivel, imod] < 0 and self.damp[ivel+1, imod] > 0):
                    if self._v > 0:
                        print('{:s}: flutter found at Mach {:.2f} for mode #{:d} near u_inf = {:.0f}.'.format(self.name, self._mach, imod, self.speed[ivel]))
                    self.f_speed = self.speed[ivel]
                    self.f_mode = imod
                    found = True
                    break
        if self._v > 0 and not found:
            print('{:s}: no flutter found at Mach {:.2f} for speed range [{:.0f}; {:.0f}].'.format(self.name, self._mach, self.speed[1], self.speed[-1]))

    def save(self, sfx):
        """Write results to disk
        sfx : filename suffix
        """
        hdr = '{:>8s}, {:>10s}, {:>10s}, {:>10s}'.format('Mach', 'u_f', 'mode', 'AGG(g)')
        np.savetxt(f'results_{sfx}.dat', np.array([self._mach, self.f_speed, self.f_mode, self.damp_agg]).reshape(1,4), fmt='%10.4f', delimiter=', ', header=hdr)
        hdr = '{:>9s}'.format('u_inf')
        for j in range(self.n_modes):
            hdr += ', {:>11s}'.format(f'f_{j}')
        for j in range(self.n_modes):
            hdr += ', {:>11s}'.format(f'g_{j}')
        data = np.concatenate((self.speed.reshape(self.speed.shape[0], 1), self.freq, self.damp), axis=1)
        np.savetxt(f'vgf_{sfx}.dat', data, fmt='%+.4e', delimiter=', ', header=hdr)

    def plot(self, sfx, show=True, format=''):
        """Plot frequency and damping as a function of the velocity
        show : whether to display the plot
        format : format to which the plot will be written to disk, if not empty
        """
        import matplotlib.pyplot as plt
        fig, axs = plt.subplots(2, 1)
        u = self.speed
        f = self.freq
        g = self.damp
        for j in range(self.n_modes):
            # Frequency
            axs[0].plot(u, f[:,j], color=f'C{j}', lw=2, ls='-')
            axs[0].set_xlabel('Velocity, $U_{\infty}$ (m/s)')
            axs[0].set_ylabel('Frequency, $f$ (Hz)')
            axs[0].spines['top'].set_visible(False)
            axs[0].spines['right'].set_visible(False)
            # Damping
            axs[1].plot(u, g[:,j], color=f'C{j}', lw=2, ls='-')
            axs[1].set_xlabel('Velocity, $U_{\infty}$ (m/s)')
            axs[1].set_ylabel('Damping, $g$')
            axs[1].spines['top'].set_visible(False)
            axs[1].spines['right'].set_visible(False)
        axs[1].plot(u, [0.] * self.n_speed, color='k', lw=1, ls='-.') # zero-damping line
        fig.tight_layout()
        if format: plt.savefig(f'vgf_{sfx}.' + format)
        if show: plt.show()
